@@ -9,7 +9,7 @@ HAS_TESTING=false
 cd ..
 for f in $(find . -name "*.go" ! -name "*_test.go" ! -name "tools.go"); do
 	if grep -q "github.com/onsi/ginkgo" $f; then
-    echo "$f imports github.com/onsi/ginkgo"
+    echo "$f imports github.com/onsi/ginkgo/v2"
     HAS_TESTING=true
 	fi
 	if grep -q "github.com/onsi/gomega" $f; then
