@@ -1,18 +1,24 @@
 package quicproxy
 
 import (
+	"errors"
+	"fmt"
 	"net"
-	"sort"
+	"os"
+	"slices"
 	"sync"
 	"time"
 
-	"github.com/lucas-clemente/quic-go/internal/protocol"
-	"github.com/lucas-clemente/quic-go/internal/utils"
+	"github.com/quic-go/quic-go/internal/protocol"
+	"github.com/quic-go/quic-go/internal/utils"
 )
 
 // Connection is a UDP connection
 type connection struct {
 	ClientAddr *net.UDPAddr // Address of the client
+	ServerAddr *net.UDPAddr // Address of the server
+
+	mx         sync.Mutex
 	ServerConn *net.UDPConn // UDP connection to server
 
 	incomingPackets chan packetEntry
@@ -25,6 +31,22 @@ func (c *connection) queuePacket(t time.Time, b []byte) {
 	c.incomingPackets <- packetEntry{Time: t, Raw: b}
 }
 
+func (c *connection) SwitchConn(conn *net.UDPConn) {
+	c.mx.Lock()
+	defer c.mx.Unlock()
+
+	old := c.ServerConn
+	old.SetReadDeadline(time.Now())
+	c.ServerConn = conn
+}
+
+func (c *connection) GetServerConn() *net.UDPConn {
+	c.mx.Lock()
+	defer c.mx.Unlock()
+
+	return c.ServerConn
+}
+
 // Direction is the direction a packet is sent.
 type Direction int
 
@@ -42,17 +64,11 @@ type packetEntry struct {
 	Raw  []byte
 }
 
-type packetEntries []packetEntry
-
-func (e packetEntries) Len() int           { return len(e) }
-func (e packetEntries) Less(i, j int) bool { return e[i].Time.Before(e[j].Time) }
-func (e packetEntries) Swap(i, j int)      { e[i], e[j] = e[j], e[i] }
-
 type queue struct {
 	sync.Mutex
 
 	timer   *utils.Timer
-	Packets packetEntries
+	Packets []packetEntry // sorted by the packetEntry.Time
 }
 
 func newQueue() *queue {
@@ -61,15 +77,27 @@ func newQueue() *queue {
 
 func (q *queue) Add(e packetEntry) {
 	q.Lock()
-	q.Packets = append(q.Packets, e)
-	if len(q.Packets) > 1 {
-		lastIndex := len(q.Packets) - 1
-		if q.Packets[lastIndex].Time.Before(q.Packets[lastIndex-1].Time) {
-			sort.Stable(q.Packets)
-		}
+	defer q.Unlock()
+
+	if len(q.Packets) == 0 {
+		q.Packets = append(q.Packets, e)
+		q.timer.Reset(e.Time)
+		return
+	}
+
+	// The packets slice is sorted by the packetEntry.Time.
+	// We only need to insert the packet at the correct position.
+	idx := slices.IndexFunc(q.Packets, func(p packetEntry) bool {
+		return p.Time.After(e.Time)
+	})
+	if idx == -1 {
+		q.Packets = append(q.Packets, e)
+	} else {
+		q.Packets = slices.Insert(q.Packets, idx, e)
+	}
+	if idx == 0 {
+		q.timer.Reset(q.Packets[0].Time)
 	}
-	q.timer.Reset(q.Packets[0].Time)
-	q.Unlock()
 }
 
 func (q *queue) Get() []byte {
@@ -113,144 +141,119 @@ func (d Direction) Is(dir Direction) bool {
 // DropCallback is a callback that determines which packet gets dropped.
 type DropCallback func(dir Direction, packet []byte) bool
 
-// NoDropper doesn't drop packets.
-var NoDropper DropCallback = func(Direction, []byte) bool {
-	return false
-}
-
 // DelayCallback is a callback that determines how much delay to apply to a packet.
 type DelayCallback func(dir Direction, packet []byte) time.Duration
 
-// NoDelay doesn't apply a delay.
-var NoDelay DelayCallback = func(Direction, []byte) time.Duration {
-	return 0
-}
+// Proxy is a QUIC proxy that can drop and delay packets.
+type Proxy struct {
+	// Conn is the UDP socket that the proxy listens on for incoming packets from clients.
+	Conn *net.UDPConn
 
-// Opts are proxy options.
-type Opts struct {
-	// The address this proxy proxies packets to.
-	RemoteAddr string
-	// DropPacket determines whether a packet gets dropped.
+	// ServerAddr is the address of the server that the proxy forwards packets to.
+	ServerAddr *net.UDPAddr
+
+	// DropPacket is a callback that determines which packet gets dropped.
 	DropPacket DropCallback
-	// DelayPacket determines how long a packet gets delayed. This allows
-	// simulating a connection with non-zero RTTs.
-	// Note that the RTT is the sum of the delay for the incoming and the outgoing packet.
-	DelayPacket DelayCallback
-}
 
-// QuicProxy is a QUIC proxy that can drop and delay packets.
-type QuicProxy struct {
-	mutex sync.Mutex
+	// DelayPacket is a callback that determines how much delay to apply to a packet.
+	DelayPacket DelayCallback
 
 	closeChan chan struct{}
+	logger    utils.Logger
 
-	conn       *net.UDPConn
-	serverAddr *net.UDPAddr
-
-	dropPacket  DropCallback
-	delayPacket DelayCallback
-
-	// Mapping from client addresses (as host:port) to connection
+	// mapping from client addresses (as host:port) to connection
+	mutex      sync.Mutex
 	clientDict map[string]*connection
-
-	logger utils.Logger
 }
 
-// NewQuicProxy creates a new UDP proxy
-func NewQuicProxy(local string, opts *Opts) (*QuicProxy, error) {
-	if opts == nil {
-		opts = &Opts{}
+func (p *Proxy) Start() error {
+	p.clientDict = make(map[string]*connection)
+	p.closeChan = make(chan struct{})
+	p.logger = utils.DefaultLogger.WithPrefix("proxy")
+
+	if err := p.Conn.SetReadBuffer(protocol.DesiredReceiveBufferSize); err != nil {
+		return err
 	}
-	laddr, err := net.ResolveUDPAddr("udp", local)
-	if err != nil {
-		return nil, err
-	}
-	conn, err := net.ListenUDP("udp", laddr)
-	if err != nil {
-		return nil, err
-	}
-	raddr, err := net.ResolveUDPAddr("udp", opts.RemoteAddr)
-	if err != nil {
-		return nil, err
+	if err := p.Conn.SetWriteBuffer(protocol.DesiredSendBufferSize); err != nil {
+		return err
 	}
 
-	packetDropper := NoDropper
-	if opts.DropPacket != nil {
-		packetDropper = opts.DropPacket
-	}
-
-	packetDelayer := NoDelay
-	if opts.DelayPacket != nil {
-		packetDelayer = opts.DelayPacket
-	}
-
-	p := QuicProxy{
-		clientDict:  make(map[string]*connection),
-		conn:        conn,
-		closeChan:   make(chan struct{}),
-		serverAddr:  raddr,
-		dropPacket:  packetDropper,
-		delayPacket: packetDelayer,
-		logger:      utils.DefaultLogger.WithPrefix("proxy"),
-	}
-
-	p.logger.Debugf("Starting UDP Proxy %s <-> %s", conn.LocalAddr(), raddr)
+	p.logger.Debugf("Starting UDP Proxy %s <-> %s", p.Conn.LocalAddr(), p.ServerAddr)
 	go p.runProxy()
-	return &p, nil
+	return nil
+}
+
+// SwitchConn switches the connection for a client,
+// identified the address that the client is sending from.
+func (p *Proxy) SwitchConn(clientAddr *net.UDPAddr, conn *net.UDPConn) error {
+	if err := conn.SetReadBuffer(protocol.DesiredReceiveBufferSize); err != nil {
+		return err
+	}
+	if err := conn.SetWriteBuffer(protocol.DesiredSendBufferSize); err != nil {
+		return err
+	}
+	p.mutex.Lock()
+	defer p.mutex.Unlock()
+	c, ok := p.clientDict[clientAddr.String()]
+	if !ok {
+		return fmt.Errorf("client %s not found", clientAddr)
+	}
+	c.SwitchConn(conn)
+	return nil
 }
 
 // Close stops the UDP Proxy
-func (p *QuicProxy) Close() error {
+func (p *Proxy) Close() error {
 	p.mutex.Lock()
 	defer p.mutex.Unlock()
+
 	close(p.closeChan)
 	for _, c := range p.clientDict {
-		if err := c.ServerConn.Close(); err != nil {
+		if err := c.GetServerConn().Close(); err != nil {
 			return err
 		}
 		c.Incoming.Close()
 		c.Outgoing.Close()
 	}
-	return p.conn.Close()
+	return nil
 }
 
 // LocalAddr is the address the proxy is listening on.
-func (p *QuicProxy) LocalAddr() net.Addr {
-	return p.conn.LocalAddr()
-}
+func (p *Proxy) LocalAddr() net.Addr { return p.Conn.LocalAddr() }
 
-// LocalPort is the UDP port number the proxy is listening on.
-func (p *QuicProxy) LocalPort() int {
-	return p.conn.LocalAddr().(*net.UDPAddr).Port
-}
-
-func (p *QuicProxy) newConnection(cliAddr *net.UDPAddr) (*connection, error) {
-	srvudp, err := net.DialUDP("udp", nil, p.serverAddr)
+func (p *Proxy) newConnection(cliAddr *net.UDPAddr) (*connection, error) {
+	conn, err := net.ListenUDP("udp", &net.UDPAddr{IP: net.IPv4(127, 0, 0, 1), Port: 0})
 	if err != nil {
 		return nil, err
 	}
+	if err := conn.SetReadBuffer(protocol.DesiredReceiveBufferSize); err != nil {
+		return nil, err
+	}
+	if err := conn.SetWriteBuffer(protocol.DesiredSendBufferSize); err != nil {
+		return nil, err
+	}
 	return &connection{
 		ClientAddr:      cliAddr,
-		ServerConn:      srvudp,
+		ServerAddr:      p.ServerAddr,
 		incomingPackets: make(chan packetEntry, 10),
 		Incoming:        newQueue(),
 		Outgoing:        newQueue(),
+		ServerConn:      conn,
 	}, nil
 }
 
 // runProxy listens on the proxy address and handles incoming packets.
-func (p *QuicProxy) runProxy() error {
+func (p *Proxy) runProxy() error {
 	for {
 		buffer := make([]byte, protocol.MaxPacketBufferSize)
-		n, cliaddr, err := p.conn.ReadFromUDP(buffer)
+		n, cliaddr, err := p.Conn.ReadFromUDP(buffer)
 		if err != nil {
 			return err
 		}
-		raw := buffer[0:n]
+		raw := buffer[:n]
 
-		saddr := cliaddr.String()
 		p.mutex.Lock()
-		conn, ok := p.clientDict[saddr]
+		conn, ok := p.clientDict[cliaddr.String()]
 
 		if !ok {
 			conn, err = p.newConnection(cliaddr)
@@ -258,31 +261,34 @@ func (p *QuicProxy) runProxy() error {
 				p.mutex.Unlock()
 				return err
 			}
-			p.clientDict[saddr] = conn
+			p.clientDict[cliaddr.String()] = conn
 			go p.runIncomingConnection(conn)
 			go p.runOutgoingConnection(conn)
 		}
 		p.mutex.Unlock()
 
-		if p.dropPacket(DirectionIncoming, raw) {
+		if p.DropPacket != nil && p.DropPacket(DirectionIncoming, raw) {
 			if p.logger.Debug() {
 				p.logger.Debugf("dropping incoming packet(%d bytes)", n)
 			}
 			continue
 		}
 
-		delay := p.delayPacket(DirectionIncoming, raw)
+		var delay time.Duration
+		if p.DelayPacket != nil {
+			delay = p.DelayPacket(DirectionIncoming, raw)
+		}
 		if delay == 0 {
 			if p.logger.Debug() {
-				p.logger.Debugf("forwarding incoming packet (%d bytes) to %s", len(raw), conn.ServerConn.RemoteAddr())
+				p.logger.Debugf("forwarding incoming packet (%d bytes) to %s", len(raw), conn.ServerAddr)
 			}
-			if _, err := conn.ServerConn.Write(raw); err != nil {
+			if _, err := conn.GetServerConn().WriteTo(raw, conn.ServerAddr); err != nil {
 				return err
 			}
 		} else {
 			now := time.Now()
 			if p.logger.Debug() {
-				p.logger.Debugf("delaying incoming packet (%d bytes) to %s by %s", len(raw), conn.ServerConn.RemoteAddr(), delay)
+				p.logger.Debugf("delaying incoming packet (%d bytes) to %s by %s", len(raw), conn.ServerAddr, delay)
 			}
 			conn.queuePacket(now.Add(delay), raw)
 		}
@@ -290,30 +296,38 @@ func (p *QuicProxy) runProxy() error {
 }
 
 // runConnection handles packets from server to a single client
-func (p *QuicProxy) runOutgoingConnection(conn *connection) error {
+func (p *Proxy) runOutgoingConnection(conn *connection) error {
 	outgoingPackets := make(chan packetEntry, 10)
 	go func() {
 		for {
 			buffer := make([]byte, protocol.MaxPacketBufferSize)
-			n, err := conn.ServerConn.Read(buffer)
+			n, err := conn.GetServerConn().Read(buffer)
 			if err != nil {
+				// when the connection is switched out, we set a deadline on the old connection,
+				// in order to return it immediately
+				if errors.Is(err, os.ErrDeadlineExceeded) {
+					continue
+				}
 				return
 			}
 			raw := buffer[0:n]
 
-			if p.dropPacket(DirectionOutgoing, raw) {
+			if p.DropPacket != nil && p.DropPacket(DirectionOutgoing, raw) {
 				if p.logger.Debug() {
 					p.logger.Debugf("dropping outgoing packet(%d bytes)", n)
 				}
 				continue
 			}
 
-			delay := p.delayPacket(DirectionOutgoing, raw)
+			var delay time.Duration
+			if p.DelayPacket != nil {
+				delay = p.DelayPacket(DirectionOutgoing, raw)
+			}
 			if delay == 0 {
 				if p.logger.Debug() {
 					p.logger.Debugf("forwarding outgoing packet (%d bytes) to %s", len(raw), conn.ClientAddr)
 				}
-				if _, err := p.conn.WriteToUDP(raw, conn.ClientAddr); err != nil {
+				if _, err := p.Conn.WriteToUDP(raw, conn.ClientAddr); err != nil {
 					return
 				}
 			} else {
@@ -334,14 +348,14 @@ func (p *QuicProxy) runOutgoingConnection(conn *connection) error {
 			conn.Outgoing.Add(e)
 		case <-conn.Outgoing.Timer():
 			conn.Outgoing.SetTimerRead()
-			if _, err := p.conn.WriteTo(conn.Outgoing.Get(), conn.ClientAddr); err != nil {
+			if _, err := p.Conn.WriteTo(conn.Outgoing.Get(), conn.ClientAddr); err != nil {
 				return err
 			}
 		}
 	}
 }
 
-func (p *QuicProxy) runIncomingConnection(conn *connection) error {
+func (p *Proxy) runIncomingConnection(conn *connection) error {
 	for {
 		select {
 		case <-p.closeChan:
@@ -351,7 +365,7 @@ func (p *QuicProxy) runIncomingConnection(conn *connection) error {
 			conn.Incoming.Add(e)
 		case <-conn.Incoming.Timer():
 			conn.Incoming.SetTimerRead()
-			if _, err := conn.ServerConn.Write(conn.Incoming.Get()); err != nil {
+			if _, err := conn.GetServerConn().WriteTo(conn.Incoming.Get(), conn.ServerAddr); err != nil {
 				return err
 			}
 		}
